# -*- coding: utf-8 -*-
"""
Anaflow subpackage providing miscellaneous tools.

Subpackages
^^^^^^^^^^^

.. currentmodule:: anaflow.tools

.. autosummary::
    laplace
    mean
    special
    coarse_graining
"""
from __future__ import absolute_import
